/*---------------------------------------------------------------------------
  This code is part of the book
  LOGIC PROGRAMMING TECHNIQUES: THE PROLOG LANGUAGE
  Ilias Sakellariou, Nikos Bassiliades, Petros Kefalas, Demosthenes Stamatis
  Hellenic Academic On-line Publications, 2015
----------------------------------------------------------------------------*/


is_valid_list([]).
is_valid_list([_Head|Tail]):-
  is_valid_list(Tail).

last_element(Last,[Last]).
last_element(Last,[_Head|Tail]):-
  last_element(Last, Tail).


%%% bits 

bit_and(0,0,0).
bit_and(0,1,0).
bit_and(1,0,0).
bit_and(1,1,1).


bitwise_and([Bit1],[Bit2],[BitR]):-
	bit_and(Bit1,Bit2,BitR).

bitwise_and([Bit1|RestBits1],[Bit2|RestBits2],[BitR|Result]):-
	bit_and(Bit1,Bit2,BitR),
	bitwise_and(RestBits1,RestBits2,Result).


%%% List Basic Operations 
	
list_member(X,[X|_]).

list_member(X,[_|Tail]):-
  list_member(X,Tail).

%%% list_length/3
%%% list_length(List,Length)
%%% Succeeds if Length is the number of elements of 
%%% list List.
list_length([],0).

list_length([_|Tail],Len):-
  list_length(Tail,LenTail),
  Len is LenTail + 1. 

%%% list_length_f/3
%%% list_length_f(List,Length)
%%% Succeeds if Length is the number of elements of 
%%% list List. The Full version
list_length_b([],0).

list_length_b([_|Tail],Len):-
  var(Len),
  list_length_b(Tail,LenTail),
  Len is LenTail + 1. 

list_length_b([_|Tail],Len):-
  number(Len), 
  Len > 0, 
  LenTail is Len - 1,
  list_length_b(Tail,LenTail).  

%%% list_delete/3
list_delete(X,[X|Tail],Tail).

list_delete(X,[Y|Tail],[Y|List]):-
     list_delete(X,Tail,List).

%%%
list_append([],List2,List2).

list_append([X|Tail],List2,[X|List3]):-
  list_append(Tail,List2,List3).


%%% list_reverse/2
list_reverse( List, RList):-
	list_reverse( List, [], RList).

list_reverse([], List, List).

list_reverse([H|T], ListSoFar, RList):-
	list_reverse(T,[H|ListSoFar], RList ).


%%%

%%% list_max/2
%%% list_max(Max,List)
list_max([X],X).
list_max([H|T],H):-
	list_max(T,MT),
	H > MT.
list_max([H|T],MT):-
	list_max(T,MT),
	H =< MT. 


%%% list_max_f/2
%%% list_max_f(Max,List)
list_max_f([X],X).
list_max_f([H|T],H):-
	list_max_f(T,MT),
	H > MT.
list_max_f([_|T],MT):-
	list_max_f(T,MT). 




list_max_b([ H |T ],Max):-
  list_max_aux( T, H, Max).

list_max_aux([],MaxSoFar,MaxSoFar).

list_max_aux( [H | T] ,MaxSoFar, Max):-
  H > MaxSoFar,
  list_max_aux( T, H, Max).

list_max_aux( [ H | T] ,MaxSoFar, Max):-
  H =< MaxSoFar,
  list_max_aux( T, MaxSoFar, Max).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
list_max_c([M],M).

list_max_c( [H1,H2|T],Max):-
  H1 > H2,
  list_max_c([H1|T],Max).

list_max_c( [H1,H2|T],Max):-
  H1 =< H2,
  list_max_c([H2|T],Max).




